import os
from requests import get
from jovian.utils.constants import API_URL
from jovian._version import __version__
from jovian.utils.logger import log
from jovian.utils.rcfile import set_notebook_slug, get_rcdata, rcfile_exists


def _u(path):
    """Make a URL from the path"""
    return API_URL + path


def _msg(res):
    try:
        data = res.json()
        if 'errors' in data and len(data['errors'] > 0):
            return data['errors'][0]['message']
        if 'message' in data:
            return data['message']
        if 'msg' in data:
            return data['msg']
    except:
        if res.text:
            return res.text
        return 'Something went wrong'


def _pretty(res):
    """Make a human readable output from an HTML response"""
    return '(HTTP ' + str(res.status_code) + ') ' + _msg(res)


def _h(fresh):
    """Create a header to provide library metadata"""
    return {"x-jovian-source": "library",
            "x-jovian-library-version": __version__,
            "x-jovian-command": "clone" if fresh else "pull"}


def get_gist(slug, fresh):
    """Download a gist"""
    res = get(url=_u('/gist/' + slug), headers=_h(fresh))
    if res.status_code == 200:
        return res.json()['data']
    raise Exception('Failed to retrieve Gist: ' + _pretty(res))


def clone(slug, fresh=True):
    """Download the files for a gist"""
    # Download gist metadata
    log('Fetching ' + slug + "..")
    gist = get_gist(slug, fresh)
    title = gist['title']

    # If fresh clone, create directory
    if fresh:
        if os.path.exists(title):
            i = 1
            while os.path.exists(title + '-' + str(i)):
                i += 1
            title = title + '-' + str(i)
        if not os.path.exists(title):
            os.makedirs(title)
        os.chdir(title)

    # Download the files
    log('Downloading files..')
    for f in gist['files']:
        with open(f['filename'], 'wb') as fp:
            fp.write(get(f['rawUrl']).content)

        # Create .jovianrc for a fresh clone
        if fresh and f['filename'].endswith('.ipynb'):
            set_notebook_slug(f['filename'], slug)

    # Print success message and instructions (TODO)
    if fresh:
        log('Cloned successfully to ' + title)
    else:
        log('Files dowloaded successfully in current directory')


RCFILE_NOTFOUND = "Could not detect '.jovianrc' file. Make sure you are running 'jovian pull' inside a directory cloned with 'jovian clone'."


def pull(slug=None):
    """Get the latest files associated with the current gist"""
    # If a slug is provided, just use that
    if slug:
        clone(slug, fresh=False)
        return

    # Check if .jovianrc exists
    if not rcfile_exists():
        log(RCFILE_NOTFOUND, error=True)
        return

    # Get list of notebooks
    nbs = get_rcdata()['notebooks']
    for fname in nbs:
        # Get the latest files for each notebook
        clone(nbs[fname]['slug'], fresh=False)



from jovian._version import __version__
from time import sleep
from jovian.utils.anaconda import upload_conda_env, CondaError
from jovian.utils.pip import upload_pip_env
from jovian.utils.api import create_gist_simple, upload_file, get_gist
from jovian.utils.logger import log
from jovian.utils.constants import WEBAPP_URL, FILENAME_MSG, RC_FILENAME
from jovian.utils.jupyter import set_notebook_name, in_notebook, save_notebook, get_notebook_name
from jovian.utils.rcfile import get_notebook_slug, set_notebook_slug, make_rcdata

set_notebook_name()

_current_slug = None


def commit(notebook_id=None, capture_env=True, filename=None, env_type='conda'):
    """Save the notebook, capture the environment, and upload to cloud for sharing"""
    global _current_slug

    # Check if we're in a Jupyter environment
    if not in_notebook():
        log('Failed to detect Juptyer notebook. Skipping..', error=True)
        return

    # Save the notebook (uses Javascript, doesn't work everywhere)
    log('Saving notebook..')
    save_notebook()
    sleep(1)

    # Get the filename of the notebook (if not provided)
    if filename is None:
        filename = get_notebook_name()

    # Exit with help message if filename wasn't detected (or provided)
    if filename is None:
        log(FILENAME_MSG)
        return

    # Check whether to create a new gist, or update an old one
    if notebook_id is None:
        # First preference to the in-memory slug variable
        if _current_slug is not None:
            notebook_id = _current_slug
        else:
            notebook_id = get_notebook_slug(filename)

    # Check if the current user can push to this slug
    if notebook_id is not None:
        gist_meta = get_gist(notebook_id)
        if not gist_meta['isOwner']:
            notebook_id = None

    # Log whether this is an update or creation
    if notebook_id is None:
        log('Creating a new notebook on https://jvn.io')
    else:
        log('Updating notebook "' + notebook_id + '" on https://jvn.io')

    # Upload the notebook & create/update the gist
    res = create_gist_simple(filename, notebook_id)
    if res is None:
        return

    # Extract slug and owner from created gist
    slug, owner = res['slug'], res['owner']

    # Set/update the slug information
    _current_slug = slug
    set_notebook_slug(filename, slug)

    # Save & upload environment
    if capture_env:
        log('Capturing environment..')
        if env_type == 'conda':
            try:
                upload_conda_env(slug)
            except CondaError as e:
                log(str(e), error=True)

    # Print commit URL
    log('Committed successfully! ' + WEBAPP_URL +
        "/" + owner['username'] + "/" + slug)



"""Utilities to read, write and manage the credentials file"""
import os
from getpass import getpass
import json
import stat
import shutil
from jovian.utils.logger import log
from jovian.utils.constants import WEBAPP_URL

CREDS = {}

HOME = os.path.expanduser('~')
CONFIG_DIR = HOME + '/.jovian'
CREDS_FNAME = 'credentials.json'
CREDS_PATH = CONFIG_DIR + '/' + CREDS_FNAME


def config_exists():
    """Check if config directory exists"""
    return os.path.exists(CONFIG_DIR)


def init_config():
    """Create the config directory"""
    if not os.path.exists(CONFIG_DIR):
        os.makedirs(CONFIG_DIR)


def purge_config():
    """Remove the config directory"""
    return shutil.rmtree(CONFIG_DIR, ignore_errors=True)


def creds_exist():
    """Check if credentials file exits"""
    return os.path.exists(CREDS_PATH)


def read_creds():
    """Read the credentials file"""
    with open(CREDS_PATH, 'r') as f:
        return json.load(f)


def write_creds(creds):
    """Write the given credentials to file"""
    init_config()
    with open(CREDS_PATH, 'w') as f:
        json.dump(creds, f)
    os.chmod(CREDS_PATH, stat.S_IREAD | stat.S_IWRITE)


def write_key(key, write_to_file=True):
    """Write the API key to memory, and the credentials file"""
    global CREDS
    CREDS['API_KEY'] = key
    if write_to_file:
        write_creds(CREDS)


def request_key():
    """Ask the user to provide the API key"""
    log("Please enter your API key (from " + WEBAPP_URL + " ):")
    api_key = getpass()
    return api_key


def read_or_request_key():
    """Read credentials file, and ask the user for API Key, if required"""
    if creds_exist():
        return read_creds()['API_KEY'], 'read'
    else:
        return request_key(), 'request'



from __future__ import print_function
from sys import stderr


def log(msg, error=False):
    """Print a message to stdout"""
    if error:
        print('[jovian] Error: ' + msg, file=stderr)
    else:
        print('[jovian] ' + msg)



WEBAPP_URL = "https://jvn.io"
API_URL = "https://api.jovian.ai"
RC_FILENAME = ".jovianrc"
FILENAME_MSG = 'Failed to detect notebook filename. Please provide the notebook filename (including .ipynb extension) as the "filename" argument to "jovian.commit".'



from requests import get, post
# from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from os.path import basename
from time import sleep
# from tqdm import tqdm, tqdm_notebook
from jovian.utils.credentials import read_or_request_key, write_key, CREDS, request_key
from jovian.utils.logger import log
from jovian.utils.jupyter import in_notebook, save_notebook, get_notebook_name
from jovian.utils.constants import API_URL
from jovian._version import __version__


class ApiError(Exception):
    """Error class for web API related Exceptions"""
    pass


def _u(path):
    """Make a URL from the path"""
    return API_URL + path


def _msg(res):
    try:
        data = res.json()
        if 'errors' in data and len(data['errors'] > 0):
            return data['errors'][0]['message']
        if 'message' in data:
            return data['message']
        if 'msg' in data:
            return data['msg']
    except:
        if res.text:
            return res.text
        return 'Something went wrong'


def _pretty(res):
    """Make a human readable output from an HTML response"""
    return '(HTTP ' + str(res.status_code) + ') ' + _msg(res)


def validate_key(key):
    """Validate the API key by making a request to server"""
    res = get(_u('/user/profile'),
              headers={'Authorization': 'Bearer ' + key})
    if res.status_code == 200:
        return True
    else:
        return False
    raise ApiError(_pretty(res))


def get_key():
    """Retrieve and validate the API Key (from memory, config or user input)"""
    if 'API_KEY' not in CREDS:
        key, source = read_or_request_key()
        if not validate_key(key):
            log('The current API key is invalid or expired.', error=True)
            key, source = request_key(), 'request'
            if not validate_key(key):
                raise ApiError('The API key provided is invalid or expired.')
        write_key(key, source == 'request')
    return CREDS['API_KEY']


def _h():
    """Create authorizaiton header with API key"""
    return {"Authorization": "Bearer " + get_key(),
            "x-jovian-source": "library",
            "x-jovian-library-version": __version__}


def get_gist(slug):
    """Get the metadata for a gist"""
    res = get(url=_u('/gist/' + slug), headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    raise Exception('Failed to retrieve metadata for notebook "' +
                    slug + '": ' + _pretty(res))


def create_gist_simple(filename=None, gist_slug=None):
    """Upload the current notebook to create a gist"""
    auth_headers = _h()

    nb_file = (filename, open(filename, 'rb'))
    log('Uploading notebook..')
    if gist_slug:
        return upload_file(gist_slug, nb_file)
    else:
        res = post(url=_u('/gist/create'),
                   data={'public': 1},
                   files={'files': nb_file},
                   headers=auth_headers)
        if res.status_code == 200:
            return res.json()['data']
        raise ApiError('File upload failed: ' + _pretty(res))


def upload_file(gist_slug, file):
    """Upload an additional file to a gist"""
    if type(file) == str:
        file = (basename(file), open('file', 'rb'))
    res = post(url=_u('/gist/' + gist_slug + '/upload'),
               files={'files': file}, headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    raise ApiError('File upload failed: ' + _pretty(res))



"""Utilies the manage the .jovianrc file"""
import os
import json
from jovian.utils.constants import RC_FILENAME


def rcfile_exists():
    """Check if .jovianrc exists"""
    return os.path.exists(RC_FILENAME)


def save_rcdata(data=None):
    """Save given data (or empty) to file"""
    if data is None:
        data = {'notebooks': {}}
    with open(RC_FILENAME, 'w') as f:
        json.dump(data, f, indent=2)


def get_rcdata():
    """Read .jovianrc file as a dict"""
    if not rcfile_exists():
        save_rcdata()
    with open(RC_FILENAME, 'r') as f:
        data = json.load(f)
    return data


def get_notebook_slug(filename):
    """Get the gist slug for a notebook filename"""
    data = get_rcdata()
    if (filename in data['notebooks']):
        return data['notebooks'][filename]['slug']


def set_notebook_slug(filename, slug):
    """Set the gist slug for a notebook filename"""
    data = get_rcdata()
    data['notebooks'][filename] = {'slug': slug}
    save_rcdata(data)


def make_rcdata(filename, slug):
    """Make an JSON string for an individual file"""
    data = {
        "notebooks": {
            filename: {
                "slug": slug
            }
        }
    }
    return json.dumps(data)
